/* sanebuttond -- SANE button monitoring daemon
   Using the SANE library.
   Copyright (C) 2004 Rene Rebe, rene@exactcode.de
   Copyright (C) 2006 Jean-Christophe Cardot, jice@cardot.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program;a if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Compile with:
   g++ `sane-config --libs -cflags` sanebuttonsd.cc  -o sanebuttonsd */

#include <assert.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/stat.h>

extern "C" {
#include <sane/sane.h>
}

#include <string>
#include <vector>
#include <iostream>

// for signal handling
#include <csignal>
#include <iostream>

static SANE_Handle device;

int   quiet  = 0;

void term(int sig)
{
  // necessary cleanup operations before terminating
  if (!quiet) std::cout << "terminating..." << std::endl;
  sane_close(device);
  exit(0);
}

void usage(void)
{
  std::cout << "Usage: sanebuttonds [--quiet] [--daemon] [--help] [--script script_name] [sane device name]" << std::endl;
  std::cout << "Options: --quiet:  produce no output" << std::endl;
  std::cout << "         --daemon: fork in the background" << std::endl;
  std::cout << "         --help:   this help page" << std::endl;
  std::cout << "         --script: run a script when a button is pressed" << std::endl;
  exit(1);
}

int main (int argc, char **argv)
{
  const SANE_Option_Descriptor *opt;
  const SANE_Device **device_list;
  SANE_Int num_dev_options = 0;

  const char *devname = 0;
  const char *defdevname = 0;
  
  SANE_Status status;
  char *full_optstring;
  SANE_Int version_code;

  /* simple argument parsing ... */
  char* script = 0;
  char* scriptarg = "";
  int   daemon = 0;
  for (argv++; *argv; argv++)
  {
    if (!strcmp(*argv, "--script")) {
      argv++;
      if (*argv) {
	script = *argv;
      }
      else {
	std::cout << "--script needs an argument!" << std::endl;
	exit (1);
      }
    }
    else if (!strcmp(*argv, "--script-arg")) {
      argv++;
      if (*argv) {
	scriptarg = *argv;
      }
      else {
	std::cout << "--script needs an argument!" << std::endl;
	exit (1);
      }
    }
    else if (!strcmp(*argv, "--daemon")) {
      daemon = 1;
    }
    else if (!strcmp(*argv, "--quiet")) {
      quiet = 1;
    }
    else if (!strcmp(*argv, "--help")) {
      usage();
    }
    else if (!devname) {
      devname = *argv;
    }
    else if ((*argv)[0] == '-' and (*argv)[1] == '-') {
      std::cout << "Unknown option: " << *argv << std::endl;
      usage();
    }
    else {
      std::cout << "Too many arguments: " << *argv << std::endl;
      usage();
    }
  }

  //FILE *f;
  if (daemon == 1) {
    int fk = -1;
    //char *fn;
    fk = fork();
    //fn = (char*)malloc(1024);
    //sprintf(fn, "/home/jice/tmp/log_%i.txt", fk);
    //f = fopen(fn, "w");
    if (fk != 0) {
	    //fputs("parent: end\n", f);
	    return 0;
    }
  }
  //char *fn = "log.txt";
  //f = fopen(fn, "w");

  //fputs("child: begin\n", f);
  //fflush(f);
  sane_init (&version_code, 0);
  
  //fputs("after sane init\n", f);
  //fflush(f);
  if (!devname)
    {
      /* If no device name was specified explicitly, we look at the
         environment variable SANE_DEFAULT_DEVICE.  If this variable
         is not set, we open the first device we find (if any): */
      devname = getenv ("SANE_DEFAULT_DEVICE");
      if (!devname)
	{
	  if (!quiet) std::cout << "Scanning for devices ..." << std::endl;
	  status = sane_get_devices (&device_list, SANE_FALSE);
	  if (status != SANE_STATUS_GOOD) {
	    std::cerr << "sane_get_devices() failed: "
		      <<  sane_strstatus (status) << std::endl;
	    return (1);
	  }
	  if (!device_list[0]) {
	    std::cerr << "no SANE devices found!" << std::endl;
	    return (1);
	  }

          std::cerr << device_list[0]->name << " " << device_list[0]->vendor << " "
	            << device_list[0]->model << " " << device_list[0]->type << std::endl;

	  devname = device_list[0]->name;
	}
    }
  
  // now we should have a valid devname
  
  status = sane_open (devname, &device);
  //fputs("after sane open\n", f);
  //fflush(f);
  if (status != SANE_STATUS_GOOD) {
    std::cerr << "open of device " << devname << " failed: "
	      << sane_strstatus (status) << std::endl;
    
    device = 0;
    return (1);
  }
  
  if (!device) {
    std::cerr << "No SANE device detected!"
	      << sane_strstatus (status) << std::endl;
    
    return (1);
  }
 
  /* now define a signal handler for cleanup */
  signal(SIGTERM, term); // register a SIGTERM handler
  // raise(SIGTERM); // will cause term() to run
 
  /* We got a device, find out how many options it has: */
  status = sane_control_option (device, 0, SANE_ACTION_GET_VALUE,
				&num_dev_options, 0);
  //fputs("after initial sane_control_option\n", f);
  //fflush(f);
  if (status != SANE_STATUS_GOOD) {
    std::cerr << "unable to determine option count" << std::endl;
    return (1);
  }
  
  bool button_pressed = false;
  std::vector<std::string> args;
  while(true) {
    if (button_pressed) {
      args.clear();

  status = sane_open (devname, &device);
  //fputs("after sane open\n", f);
  //fflush(f);
  if (status != SANE_STATUS_GOOD) {
    std::cerr << "open of device " << devname << " failed: "
	      << sane_strstatus (status) << std::endl;
    
    device = 0;
    return (1);
  }
  
  if (!device) {
    std::cerr << "No SANE device detected!"
	      << sane_strstatus (status) << std::endl;
    
    return (1);
  }
 
  /* We got a device, find out how many options it has: */
  status = sane_control_option (device, 0, SANE_ACTION_GET_VALUE,
				&num_dev_options, 0);
  //fputs("after initial sane_control_option\n", f);
  //fflush(f);
  if (status != SANE_STATUS_GOOD) {
    std::cerr << "unable to determine option count" << std::endl;
    return (1);
  }
    }
    button_pressed = false;
  while (script && !button_pressed || !script)
    {
      if (!script) /* if no script we run in demo mode */
	button_pressed = false;
      for (int i = 0; i < num_dev_options; ++i)
	{
	  opt = 0;
	  opt = sane_get_option_descriptor (device, i);
  //fputs("after sane_get_option_descriptor\n", f);
  //fflush(f);
	  
	  if (opt->name && SANE_OPTION_IS_ACTIVE(opt->cap)) {
	    std::string optname (opt->name);
  //fputs("option name: ", f);
  //fputs(optname.c_str(), f);
  //fputs("\n", f);
  //fflush(f);
	    if (optname.find("button") == 0 || optname == "message") {
  //fputs("this is a button!\n", f);
  //fflush(f);
	      SANE_Int v_i = 0;
	      char v_s[1024]; v_s[0] = 0;
	      
	      switch (opt->type)
		{
		case SANE_TYPE_BOOL:
                case SANE_TYPE_INT:
		  status = sane_control_option (device, i,
		                                SANE_ACTION_GET_VALUE,
                                                &v_i, 0);
		  break;
		case SANE_TYPE_FIXED:
		  break;
		case SANE_TYPE_STRING:
                  status = sane_control_option (device, i,
                                                SANE_ACTION_GET_VALUE,
                                                v_s, 0);
		  break;
		case SANE_TYPE_BUTTON:
		  break;
		case SANE_TYPE_GROUP:
		  break;
		}

  //fputs("after switch\n", f);
  //fflush(f);
              if (status != SANE_STATUS_GOOD) {
  //fputs("unable to obtain option value\n", f);
  //fflush(f);
                std::cerr << "unable to obtain option value: "
                          << sane_strstatus (status) << std::endl;
		continue;
              }
  //fputs("status == SANE_STATUS_GOOD\n", f);
  //fflush(f);

	      if (opt->type == SANE_TYPE_BOOL && v_i) {
		if (!quiet) std::cout << opt->title << std::endl;
		std::string s (opt->name);
  //fputs("s=", f);
  //fputs(s.c_str(), f);
  //fputs("\n", f);
  //fflush(f);
		int i = s.find(' ');
		s.erase (0, i + 1);
		s.insert(0, "-");
  //fputs("s=", f);
  //fputs(s.c_str(), f);
  //fputs("\n", f);
  //fflush(f);
		args.push_back (s);
		button_pressed = true;
  //fputs("button pressed = true\n", f);
  //fflush(f);
	       }
	      else if (opt->type == SANE_TYPE_STRING && v_s[0] != 0 &&
		       button_pressed) {
		if (!quiet) std::cout << "Message " << v_s << std::endl;
		args.push_back (v_s);
	      }
	    } // if button
	  } // if active
	} // for each option
  //fputs("options finies de parcourir\n", f);
  //fflush(f);
      if (!button_pressed)
	usleep (250000);
    } // while (true)
  //fputs("after button detection\n", f);
  //fflush(f);
  sane_close (device);
  //fputs("after sane close\n", f);
  //fflush(f);

  if (script && button_pressed) {
    const char* c_args[10];
    const char** ca = c_args; *ca++ = script;
    char *cargs;
    cargs = (char*)malloc(1024);
    *cargs = 0;
    for (std::vector<std::string>::iterator it = args.begin();
         it != args.end(); ++it) {
      *ca++ = it->c_str();
      sprintf(cargs, "%s %s", cargs, it->c_str());
    }
    *ca = NULL;
    //fputs("just before exec\n", f);
    //fflush(f);
    //execv(script, (char* const*)c_args);
    char *syscde;
    syscde = (char*)malloc(1024);
    sprintf(syscde, "%s %s %s", script, scriptarg, cargs);
    //fputs(syscde, f);
    //fputs("\n", f);
    //fflush(f);
    system(syscde);
    //fputs("after exec\n", f);
    //fflush(f);
  }
  } // while(true)
  return 0;
}

